#!/usr/bin/env python3

import singer
from singer.catalog import Catalog, CatalogEntry, Schema
from singer.utils import parse_args
from singer import metadata

from . import streams
from .context import Context
from .http import authed_get
from .streams import Scopes
from .syncs import Syncer

LOGGER = singer.get_logger()


REQUIRED_CONFIG_KEYS = ["start_date", 'api_key']


def check_credentials_are_authorized(ctx):
    res = authed_get(Scopes.source, Scopes.endpoint, ctx.config)
    scopes = res.json().get('scopes', [])

    missing_auths = set(Scopes.scopes)
    for s in Scopes.scopes:
        if s in scopes:
            missing_auths.remove(s)

    if len(missing_auths):
        raise Exception('Insufficient authorization, missing for {}'.format(
            ','.join(missing_auths)
        ))


def discover(ctx):
    check_credentials_are_authorized(ctx)
    catalog = Catalog([])
    for stream in streams.STREAMS:
        schema = Schema.from_dict(streams.load_schema(stream.tap_stream_id),
                                  inclusion="available")

        mdata = metadata.new()

        for prop in schema.properties:
            if prop in streams.PK_FIELDS[stream.tap_stream_id]:
                mdata = metadata.write(mdata, ('properties', prop), 'inclusion', 'automatic')
            else:
                mdata = metadata.write(mdata, ('properties', prop), 'inclusion', 'available')

        catalog.streams.append(CatalogEntry(
            stream=stream.tap_stream_id,
            tap_stream_id=stream.tap_stream_id,
            key_properties=streams.PK_FIELDS[stream.tap_stream_id],
            schema=schema,
            metadata=metadata.to_list(mdata)
        ))
    return catalog


def desired_fields(selected, stream_schema):
    '''
    Returns fields that should be synced
    '''
    all_fields = set()
    available = set()
    automatic = set()

    for field, field_schema in stream_schema.properties.items():
        all_fields.add(field)
        inclusion = field_schema.inclusion
        if inclusion == 'automatic':
            automatic.add(field)
        elif inclusion == 'available':
            available.add(field)
        else:
            raise Exception('Unknown inclusion ' + inclusion)

    not_selected_but_automatic = automatic.difference(selected)
    if not_selected_but_automatic:
        LOGGER.warning(
            'Fields %s are required but were not selected. Adding them.',
            not_selected_but_automatic)

    return selected.intersection(available).union(automatic)


def sync(ctx):
    check_credentials_are_authorized(ctx)

    for c in ctx.selected_catalog:
        selected_fields = set(
            [k for k, v in c.schema.properties.items()
             if v.selected or k == c.replication_key])
        fields = desired_fields(selected_fields, c.schema)

        schema = Schema(
            type='object',
            properties={prop: c.schema.properties[prop] for prop in fields}
        )
        c.schema = schema
        streams.write_schema(c.tap_stream_id, schema)

    syncer = Syncer(ctx)
    syncer.sync()


def main_impl():
    args = parse_args(REQUIRED_CONFIG_KEYS)
    ctx = Context(args.config, args.state)
    if args.discover:
        discover(ctx).dump()
    else:
        ctx.catalog = Catalog.from_dict(args.properties) \
            if args.properties else discover(ctx)
        sync(ctx)


def main():
    try:
        main_impl()
    except Exception as exc:
        LOGGER.critical(exc)
        raise

if __name__ == '__main__':
    main()


import singer

from .streams import PK_FIELDS, STREAMS, IDS


def get_results_from_payload(payload):
    """
    SG sometimes returns Lists or one keyed Dicts
    """
    if isinstance(payload, dict):
        return next(iter(payload.values()))

    else:
        return payload


def make_record_if_str(record, stream):
    """
    transform email string to dict for group suppression members
    """
    if isinstance(record, str):
        record = {PK_FIELDS[stream.tap_stream_id][0]: record}

    return record


def send_selected_properties(schema, record, stream, added_properties):
    """
    Creates and returns new record with selected properties
    """
    r = make_record_if_str(record, stream)

    record = {
        field: r.get(field) for field, val
        in schema.to_dict()['properties'].items() if val['selected'] or val['inclusion'] == 'automatic'
    }

    if added_properties:
        record.update(added_properties)

    return record


def trimmed_records(schema, data, stream, added_properties):
    """
    Takes raw data and details on what to sync and returns cleaned to records
    with only selected fields
    """
    return [send_selected_properties(schema, r, stream, added_properties)
            for r in data]


def get_added_properties(stream, id):
    return {"%s_id" % stream.tap_stream_id.split('_')[0][:-1]: id}


def trim_members_all(tap_stream_id):
    """
    E.g. returns groups for groups_all
    """
    return tap_stream_id.split('_')[0]


def add_all(tap_stream_id):
    """
    Adds all to the generic term e.g. groups_all for groups
    """
    return tap_stream_id.split('-')[0] + '_all'


def find_old_list_count(list_id, all_lists_state):
    """
    Returns the last list size saved for the provided list
    :param list_id:
    :param all_lists_state:
    :return:
    """
    last_size = 0
    for x in all_lists_state:
        if x['id'] == list_id:
            last_size = x['member_count']

    return last_size


def clean_for_cache(data, tap_stream_id):
    """
    For saving lists sizes to cache, clean to just ID and member count.
    Applicable to GROUPS, LISTS, and SEGMENTS
    """
    lookup_keys = {
        IDS.LISTS_ALL: 'recipient_count',
        IDS.GROUPS_ALL: 'unsubscribes',
        IDS.SEGMENTS_ALL: 'recipient_count',
    }
    if tap_stream_id in lookup_keys:
        return [
            {
                'id': d['id'],
                'member_count': d[lookup_keys[tap_stream_id]]
            } for d in data
        ]
    else:
        return data


def safe_update_dict(obj1, obj2):
    if obj2:
        obj1.update(obj2)
    return obj1


def get_tap_stream_tuple(tap_stream_id):
    for s in STREAMS:
        if s.tap_stream_id == tap_stream_id:
            return s


def write_metrics(tap_stream_id, records):
    with singer.metrics.record_counter(tap_stream_id) as counter:
        counter.increment(len(records))


def write_records(tap_stream_id, records):
    singer.write_records(tap_stream_id, records)
    write_metrics(tap_stream_id, records)


import singer
from .streams import IDS
from .http import authed_get, end_of_records_check
from .utils import (
    trimmed_records, trim_members_all, get_results_from_payload,
    safe_update_dict,
    write_records, get_tap_stream_tuple, find_old_list_count,
    get_added_properties
)

logger = singer.get_logger()


class Syncer(object):

    def __init__(self, ctx):
        self.ctx = ctx

    def sync(self):
        self.sync_alls()
        self.sync_incrementals()
        self.ctx.write_state()

    def sync_incrementals(self):
        for cat_entry in self.ctx.selected_catalog:
            stream = get_tap_stream_tuple(cat_entry.tap_stream_id)
            if stream.bookmark:
                getattr(self, 'sync_%s' % stream.bookmark[1])(
                    stream, cat_entry.schema)

                self.ctx.write_state()

    def sync_timestamp(self, stream, schema):
        """
        Searches using created and updated ates as created doesn't also
        impact updated
        """
        start = self.ctx.update_start_date_bookmark(stream.bookmark)

        for day in self.discrete_days_since_start(start):
            for search_term in ['created_at', 'updated_at']:
                params = {
                    search_term: day.int_timestamp
                }
                logger.info("Looking for contacts %s on %s" % (
                    search_term, day.to_date_string()))
                self.write_paged_records(stream, schema, params=params)
            self.ctx.set_bookmark(stream.bookmark, self.ctx.now_date_str())

    def write_paged_records(self, stream, schema,
                            params=None,
                            url_key=None,
                            added_properties=None):

        for res in self.get_using_paged(stream, add_params=params, url_key=url_key):
            results = res.json().get('recipients')
            if results:
                self.write_records(schema, results, stream,
                                   added_properties=added_properties)

    @staticmethod
    def write_records(schema, results, stream, added_properties=None):
        records = trimmed_records(schema, results, stream, added_properties)
        write_records(stream.tap_stream_id, records)

    def sync_end_time(self, stream, schema):
        start = self.ctx.update_start_date_bookmark(stream.bookmark).int_timestamp
        end = self.ctx.now_seconds

        logger.info('Starting to extract %s from %s to %s' % (
            stream.tap_stream_id, str(start), str(end)))

        for results in self.get_using_offset(stream, start, end):
            self.write_records(schema, results, stream)
            self.ctx.set_bookmark(stream.bookmark, self.ctx.ts_to_dt(end))

    def sync_member_count(self, stream, schema):
        stream_type = trim_members_all(stream.tap_stream_id)

        for list in self.ctx.cache[stream_type]:
            old_list_count = find_old_list_count(
                list['id'],
                self.ctx.update_start_date_bookmark(stream.bookmark))
            if list['member_count'] > old_list_count:
                logger.info('Starting to extract %s as list size now: %s, was: %s' % (
                    stream.tap_stream_id, list['member_count'], old_list_count))

                self.get_and_write_members(list, stream, schema)
            else:
                logger.info('Not syncing %s %s as it is same size as last sync'
                            % (stream_type, list['id']))

    def sync_alls(self):
        for cat_entry in self.ctx.selected_catalog:
            stream = get_tap_stream_tuple(cat_entry.tap_stream_id)
            if not stream.bookmark:
                logger.info('Extracting all %s' % stream.tap_stream_id)

                results = self.get_alls(stream)
                self.write_records(cat_entry.schema, results, stream)
                self.ctx.update_cache(results, cat_entry.tap_stream_id)

    def get_and_write_members(self, list, stream, schema):
        added_properties = get_added_properties(stream, list['id'])

        if stream.tap_stream_id == IDS.GROUPS_MEMBERS:
            results = self.get_alls(stream, url_key=list['id'])
            self.write_records(schema, results, stream,
                               added_properties=added_properties)

        else:
            self.write_paged_records(stream, schema, url_key=list['id'],
                                     added_properties=added_properties)

        self.ctx.save_member_count_state(list, stream)

    def get_alls(self, stream, url_key=None):
        endpoint = stream.endpoint.format(url_key) if url_key else stream.endpoint

        return get_results_from_payload(authed_get(
            stream.tap_stream_id, endpoint, self.ctx.config).json())

    def get_using_paged(self, stream, add_params=None, url_key=None):
        page = 1
        page_size = 1000
        endpoint = stream.endpoint.format(url_key) if url_key else stream.endpoint

        while True:
            params = {
                'page': page,
                'page_size': page_size
            }
            safe_update_dict(params, add_params)
            r = authed_get(stream.tap_stream_id,
                           endpoint,
                           self.ctx.config,
                           params=params)
            yield r
            if not end_of_records_check(r):
                page += 1
            else:
                break

    def get_using_offset(self, stream, start, end):
        offset = 0
        limit = 500

        while True:
            r = authed_get(
                stream.tap_stream_id,
                stream.endpoint,
                self.ctx.config,
                params={
                    'offset': offset,
                    'limit': limit,
                    'start_time': start,
                    'end_time': end
                }
            )
            yield r.json()
            if len(r.json()):
                offset += limit
            else:
                break

    def discrete_days_since_start(self, start):
        """
        Return List timestamps each day since start
        """
        search_term = start.start_of('day')
        days_to_search = []
        while search_term <= self.ctx.now:
            days_to_search.append(search_term)
            search_term = search_term.add(days=1)

        return days_to_search



import requests
from singer import metrics

session = requests.Session()


def authed_get(tap_stream_id, url, config, params=None):
    headers = {"Authorization": "Bearer %s" % config['api_key']}
    with metrics.http_request_timer(tap_stream_id) as timer:
        resp = session.request(method='get', url=url, params=params, headers=headers)
        timer.tags[metrics.Tag.http_status_code] = resp.status_code
        return resp


def end_of_records_check(r):
    empty_message = "No more pages"
    if r.status_code == 404 and r.json().get(
            'errors', [{}])[0].get('message') == empty_message:
        return True
    if r.json().get('recipient_count') == 0:
        return True
    else:
        return False
